#include "aap.h"
#include "aap_aclm.h"
#include "aap_aiu.h"
#include "aap_asa.h"
#include "aap_ana.h"
#include "aap_aitu.h"
#include "aap_asrp.h"
#include "claimer.h"

#include <stdio.h>
#include <stdlib.h>
/*------------------------------------------------------------------*/
static void remove_addr_claim(aap *aap, addr_range *ar);
static int aap_receive(void *obj, char *data, int len, net_peer *np);
static int aap_timeout(void *obj, unsigned int time);
static void startup(aap *a);
static int next_rseq(aap *aap);
/*------------------------------------------------------------------*/
struct aap_allocation {
    aap *aap;
    int numallocs;
    unsigned int end_time;
    aap_callback_func f;
    void *obj;
};

/*------------------------------------------------------------------*/
static int forward_allocate(void *obj, unsigned int time) {
    claimer *clm;
    struct aap_allocation *aa = obj;
  
    clm = claimer_new(aa->aap, aa->numallocs, aa->end_time, aa->f, aa->obj);
    clm->claim(clm);
    aa->aap->claim_list->insert(aa->aap->claim_list, clm);
    return 0;
}


/*------------------------------------------------------------------*/
static void allocate_addr(aap *aap, int numallocs, unsigned int end_time,
                          aap_callback_func f, void *obj){
    claimer *clm;
    if(aap->state == AAP_RUNNING) {
        clm = claimer_new(aap, numallocs, end_time, f, obj);
        clm->claim(clm);
        aap->claim_list->insert(aap->claim_list, clm);

    } else {
        struct aap_allocation *aa = malloc(sizeof(struct aap_allocation));
        aa->aap = aap;
        aa->numallocs = numallocs;
        aa->end_time = end_time;
        aa->f = f;
        aa->obj = obj;
  
        aap->tim->set(aap->tim, aap->cfg->STARTUP_WAIT - (time(NULL) - aap->startup_time),
                      forward_allocate, aa);
    }
}
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/

static void remove_addr_claim(aap *aap, addr_range *ar){

    int i;
  
    for(i = 0 ; i < aap->claim_list->length(aap->claim_list) ; i++){
        addr_range *tmp_ar;
        tmp_ar = aap->claim_list->elem_at(aap->claim_list, i);
    
        /* if the address_range overlaps, remove the
           entry from unanswered_aclm list */
        if( tmp_ar->overlap(tmp_ar,ar) ){
            aap->claim_list->remove(aap->claim_list, i);
      
        }
    } 
  
}

/*------------------------------------------------------------------*/
/*
 * startup sets the startup timer, and binds and sets the destination
 * address for the net_socket associated with the aap object.
 *
 */
static void startup(aap *a){
  
    int startup_wait = ((double)(rand())/RAND_MAX * 0.3 + 1) * a->cfg->STARTUP_WAIT; 

    a->in_ns->bind(a->in_ns, a->cfg->aap_addr, a->cfg->aap_port);
    a->out_ns->set_dest(a->out_ns, a->cfg->aap_addr, a->cfg->aap_port);
  
    a->net->receive(a->net, a->in_ns, aap_receive, a);
    a->tim->set(a->tim, startup_wait, aap_timeout, a);
}
/*------------------------------------------------------------------*/
/*
 * periodic_aiu is the timer callback function for sending the
 * periodic AIU messages. It constructs and sends an AIU message
 * and sets a timer for the next periodic message.
 *
 */
static int periodic_aiu(void *obj, unsigned int time) {
    int repeat_interval;
    aap *a = obj;
    aap_aiu *aiu = aap_aiu_new(a->net, a->out_ns, a->tim, a->cfg, a,
                               a->arec->allocations(a->arec), next_rseq(a));
    aiu->send(aiu);
    aiu->delete(aiu);

    /* repeat_interval is REPEAT_INTERVAL +-30% to avoid sync effects */
    repeat_interval = a->cfg->REPEAT_INTERVAL +
        (((double)rand()/RAND_MAX) * 0.6 - 0.3) * a->cfg->REPEAT_INTERVAL;
  
    a->tim->set(a->tim, repeat_interval, periodic_aiu, a);
    return 0;
}
/*------------------------------------------------------------------*/
/*
 * periodic_asa is the timer callback function for sending the
 * periodic ASA messages. It constructs and sends an ASA message
 * and sets a timer for the next periodic message.
 */
static int periodic_asa(void *obj, unsigned int time) {
    int repeat_interval;
    list *ranges;
    aap *a = obj;
    aap_asa *asa = aap_asa_new(a->net, a->out_ns, a->tim, a->cfg, a, next_rseq(a));
  
    ranges = list_new();
  
    /* insert the defined address scope */
    ranges->insert(ranges, a->cfg->scope);

    asa->send(asa, ranges);
    asa->delete(asa);

    /* repeat_interval is REPEAT_INTERVAL +-30% to avoid sync effects */
    repeat_interval = a->cfg->ASA_INTERVAL +
        (((double)rand()/RAND_MAX) * 0.6 - 0.3) * a->cfg->ASA_INTERVAL;
  
    a->tim->set(a->tim, repeat_interval, periodic_asa, a);
    return 0;
}

static int periodic_asrp(void *obj, unsigned int time){
    aap *a = obj;
    aap_asrp *asrp = aap_asrp_new(a->net, a->out_ns, a->tim, a->cfg, a, next_rseq(a));
    int asrp_timeout = a->cfg->ASRP_INTERVAL +
        (((double)rand()/RAND_MAX) * 0.6 - 0.3) * a->cfg->ASRP_INTERVAL;
    asrp->send(asrp);
    asrp->delete(asrp);
    a->asrp_timer = a->tim->set(a->tim, asrp_timeout, periodic_asrp, a);
    return 0;
}

/*------------------------------------------------------------------*/
/*
 * aap_receive is the callback function that is called whenever data
 * is received on the net_socket associated with the app object. The
 * data is parsed as an AAP message, and dispatched to the appropriate
 * message handler (e.g., aiu_receive for AIU messages).
 *
 */
static int aap_receive(void *obj, char *data, int len, net_peer *np){

    aap *a = obj;
    aap_msg *am = aap_msg_new();


    if( am->parse(am, data, len, np) == -1 ){
        /* erroneous message */
        printf("aap_receive: erroneous message\n");
        return 1;
    }
  
    am->debug_print(am);

    switch(am->msgtype(am)) {
    case AAP_MSG_ACLM:
        if(a->cfg->type != CFG_PC) {
            aap_aclm_receive(a, am);
        }
        break;
    case AAP_MSG_AIU:
        if(a->cfg->type != CFG_PC) {
            aap_aiu_receive(a, am);
        }
        break;
    case AAP_MSG_AITU:
        if(a->cfg->type != CFG_PC) {
            aap_aitu_receive(a, am);
        }
        break;
    case AAP_MSG_ASA:
        aap_asa_receive(a, am);    
        break;
    case AAP_MSG_ASRP:
        aap_asrp_receive(a, am);
        break;    
    case AAP_MSG_ANA:
        if(a->cfg->type != CFG_PC) {
            aap_ana_receive(a, am);
        }
        break;
    default:
        /* obscure message received */
        break;
    }


    return 0;
}
/*------------------------------------------------------------------*/
/*
 * aap_timeout is the timer callback set during startup. It merely
 * sets AAP into running mode, so that it can start sending messages.
 *
 */
static int aap_timeout(void *obj, unsigned int time){
    aap *a = obj;
    int asrp_timeout = a->cfg->ASRP_INTERVAL +
        (((double)rand()/RAND_MAX) * 0.6 - 0.3) * a->cfg->ASRP_INTERVAL;
    /* if large scope AND ASA msg received*/
    a->state = AAP_RUNNING;

    printf("AAP startup timeout\n");
    a->tim->set(a->tim, 0, periodic_aiu, a);
    if(a->cfg->scope == CFG_LARGE_SCOPE){
        a->asrp_timer = a->tim->set(a->tim, asrp_timeout, periodic_asrp, a);
    }

    if( a->cfg->type == CFG_PC ){

        a->tim->set(a->tim, 0, periodic_asa, a);
    
    }
  
    return 0;
   
}
/*------------------------------------------------------------------*/
static int next_rseq(aap *aap) {
    return aap->rseq++;
}
/*------------------------------------------------------------------*/
aap *aap_new(net *n, tim *t, cfg *c){

    aap *a;
 
    a = (aap *)malloc(sizeof(aap));

    a->state = AAP_STARTUP;
    a->net = n;
    a->tim = t;
    a->cfg = c;
    a->claim_list = list_new();
    a->defens_list = list_new();

    a->arec = arec_new(c, t, c->scope);

    a->asa_exp_timer = 0;
    a->asa_resend_timer = 0;
    a->last_asa_exp_time= 0;
    a->last_asa_receive_time = 0;
    a->last_asa_body = (char *)calloc(508, sizeof(char));
    a->last_asa_body_len = 0;
  
    a->in_ns = n->socket(n);
    a->out_ns = n->socket(n);
    /*  a->out_ns = a->in_ns;*/


    a->rseq = 0;

    a->startup_time = time(NULL);
  
    a->allocate = allocate_addr;
  
    /* bizarre casts needed since aap_aclm.h cannot be included in aap.h
       due to nesting problems */
    /*  a->remove_claim = (void (*)(aap *, addr_range *))remove_addr_claim;*/
    /*  a->find_claim = (void *(*)(aap *, addr_range *))find_addr_claim;*/
    a->start = startup;

    a->next_rseq = next_rseq;
    

    return a;

}
/*------------------------------------------------------------------*/
