#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include "aap_aclm.h"
#include "aap_aiu.h"
#include "claimer.h"
#include "defender.h"

/*------------------------------------------------------------------*/
struct aiu_raw {
    unsigned int current_time;
    addr_range_raw ranges[500];
};

/*------------------------------------------------------------------*/
static int send_msg(aap_aiu *a) {

    aap_msg *msg;
    struct aiu_raw aiu;
    int len, i;
    list *allocs;
  
    /* construct the message body */
    bzero(&aiu, sizeof(struct aiu_raw));

    /* insert current time */
    aiu.current_time = htonl(time(NULL));

    allocs = a->allocations;

    /* copy address ranges into message body */
    for(i = 0; i < allocs->length(allocs); i++) {
        addr_range *ar;
        allocation *alloc = allocs->elem_at(allocs, i);
        ar = alloc->range(alloc);
        aiu.ranges[i] = ar->raw(ar);
    }
    /* total length of the message body */
    len = i * sizeof(addr_range_raw) + 4;
  
    /* only send message if there are allocations */
    if(i > 0) {
        msg = aap_msg_new();
        msg->create(msg, AAP_MSG_AIU, AAP_MSG_IP,
                    a->rseq, a->mseq, (char *)&aiu, len);
        msg->send(msg, a->n, a->ns);
        a->mseq++;
    }
    return 0;
}

/*------------------------------------------------------------------*/

int aap_aiu_receive(aap *aap, aap_msg *msg) {
    struct aiu_raw *aiu;
    unsigned int clock_skew;
    int rseq, mseq, addrfamily;
  
    int len, i, j;
    defender *def;
    claimer *clm;
    addr_range *ar, *clm_ar, *def_ar;
    allocation *alloc;
    list *overlaps;  

  
    rseq = msg->rseq(msg);
    mseq = msg->mseq(msg);
    addrfamily = msg->addrfamily(msg);
    
    msg->body(msg, (char **)&aiu, &len);

    /* adjust for clock skew as in section
       5.3.1 of draft-ietf-malloc-aap-03.txt */
    clock_skew = ntohl(aiu->current_time) - time(NULL);
  
    /* i should contain the number of ranges */
    i = (len - 4) / sizeof(addr_range_raw);
    if(len != (i * sizeof(addr_range_raw) + 4)) {
        /* malformed message */
        DEBUG('d', "aap_aiu: Received malformed message.\n");
        return 1;
    }
 
    /* process the address ranges */
    /* check if the addr_range overlaps an existing allocation */
    /* check if there are outstanding ACLMs to discard if overlaps*/
    /* update Allocation record */
    /* check unanswered_aclm to see if the address is claimed */
    
    while(i > 0) {
    
        ar = addr_range_new_from_raw(&(aiu->ranges[i - 1]));
        ar->adjust(ar, clock_skew);
    
        len = aap->defens_list->length(aap->defens_list);
    
        for(j = 0; j < len; j++) {
            def = aap->defens_list->elem_at(aap->defens_list, j);
            def_ar = def->range(def);
            printf("Defending ");
            def_ar->debug_print(def_ar);
            if(ar->overlap(ar, def_ar)) {
                def->double_interval(def);
            }
        }


        /* check if we have any ACLMs to discard, and do so */
        overlaps = aap->arec->overlap_claims(aap->arec, ar);
        for(i = 0; i < overlaps->length(overlaps); i++) {
            aap->arec->remove_claim(aap->arec, overlaps->elem_at(overlaps, i));
        }
        overlaps->delete(overlaps);

    
        /* update Allocation record, but first check if it overlaps an existing allocation */
        overlaps = aap->arec->overlap(aap->arec, ar);
        if( overlaps->length(overlaps) > 0 ){
            /* for each overlapping addr_range, check to see if
               it is a update and is from the same MAAS */
            
            for(j = 0; j < overlaps->length(overlaps) ; j++){

                allocation *tmp_alloc = overlaps->elem_at(overlaps, j);
	
                if((tmp_alloc->owner_addr(tmp_alloc) == msg->sender(msg)) ||
                   (tmp_alloc->flag == ALLOCATION_PREALLOC)){
                    /* update the end time for the overlaped allocation */	
		    	  	  
                    aap->arec->remove(aap->arec, tmp_alloc);	  
                    tmp_alloc->delete(&tmp_alloc);

                    alloc = allocation_new(ALLOCATION_ALLOC, ar, 0, msg->sender(msg));
                    /*	  printf("sender: %d %d\n",tmp_alloc->owner_addr(tmp_alloc), msg->sender(msg));*/
                    aap->arec->insert(aap->arec, alloc);	  

                } else {
                    /* was aiu was not sent from the owner of the allocation */
                    /* do nothing, just ignore it */
                }

            }

        } else {
            /* the allocation is inserted into the allocation record */

            alloc = allocation_new(0, ar, 0, msg->sender(msg));       
            aap->arec->insert(aap->arec, alloc);
        }
        overlaps->delete(overlaps);      

    
        len = aap->claim_list->length(aap->claim_list);
        for(j = 0 ; j < len; j++){
            clm = aap->claim_list->elem_at(aap->claim_list, j);
            clm_ar = clm->range(clm);
            /* if the address_range overlaps, remove the
               entry from unanswered_aclm list */

            if(ar->overlap(ar, clm_ar)) {
                clm->retry(clm);
            }
        }




        i--;
    }
  
    return 0;

}
/*------------------------------------------------------------------*/
static void delete(aap_aiu *a) {
    free(a);
}
/*------------------------------------------------------------------*/
aap_aiu *aap_aiu_new(net *n, net_socket *ns, tim *t, cfg *c, aap *aap, list *allocs, int rseq) {
  
    aap_aiu *aiu = malloc(sizeof(aap_aiu));

    aiu->n = n;
    aiu->ns = ns;
    aiu->t = t;
    aiu->c = c;
    aiu->aap = aap;
    aiu->allocations = allocs;
    aiu->rseq = rseq;
    aiu->mseq = 0;
  
    aiu->send = send_msg;
    aiu->delete = delete;
  

    return aiu;
}
/*------------------------------------------------------------------*/
