#include <stdio.h>
#include <stdlib.h>

#include "aap_aclm.h"
#include "aap_ana.h"
#include "claimer.h"

static int aiu_expire_callback(void *obj, unsigned int time);
static int aiu_resend_callback(void *obj, unsigned int time);

static void delete(claimer *clm);
/*
 * resend_callback is the timer callback set in order to resend
 * the ACLM message with exponential intervals.
 *
 */
static int resend_callback(void *obj, unsigned int time) {
    claimer *clm = obj;
    aap_aclm *a = clm->aclm;

    /* send the ACLM message */
    a->send(a);
  
    /* double the interval to next resend */
    clm->last_interval = clm->last_interval * 2;  
    clm->resend_timer = clm->t->set(clm->t, clm->last_interval, resend_callback, clm);
    return 0;
}
/*------------------------------------------------------------------*/
/*
 * claim_callback is the timer callback that is called when a claim
 * has succeeded, i.e., when the Claim timer expires. It cancels
 * the resend timer.
 *
 */
static int claim_callback(void *obj, unsigned int time) {
    int pos;
    claimer *clm = obj;
    aap_aclm *a = clm->aclm;

    list *alloc_list = list_new();
    
    /* cancel the resend timer */
    clm->t->unset(clm->t, clm->resend_timer);

    /* remove from claim_list */
    pos = clm->aap->claim_list->find(clm->aap->claim_list, clm);

    if( pos != -1 ){
        clm->aap->claim_list->remove(clm->aap->claim_list, pos);
    }

    clm->cancel(clm);
  
    /* create an AIU message */ 
    alloc_list->insert(alloc_list, clm->alloc);
    clm->aiu = aap_aiu_new(clm->n, clm->ns, clm->t, a->cfg, clm->aap, alloc_list,
                           clm->aap->next_rseq(clm->aap));

    /* set timer for the AIU */
    clm->claim_timer = clm->t->set(clm->t, clm->cfg->REPEAT_INTERVAL, aiu_expire_callback, clm);
  
    /* set timer for resending AIU */
    clm->last_interval = clm->cfg->RESEND_WAIT;
    clm->resend_timer = clm->t->set(clm->t, clm->last_interval, aiu_resend_callback , clm);

    clm->aclm->delete(clm->aclm);
    clm->aclm = NULL;
    
    /* insert the allocated addresses into Allocation record */ 
    clm->aap->arec->insert_own(clm->aap->arec, clm->alloc);
  
    /* send the first AIU */
    clm->aiu->send(clm->aiu);

    /* call the callback function associated with this claim,
       announcing the address range we have claimed */
    clm->f(clm->obj, clm->range(clm));

    return 0;
}

/*------------------------------------------------------------------*/
/**
 * This callback function is called when it is time to
 * stop resending separate AIUs.
 */
static int aiu_expire_callback(void *obj, unsigned int time){
    claimer *clm = obj;
  
    /* remove the last set resend timer */
    clm->t->unset(clm->t, clm->resend_timer);

    /* erase this ACLM */
    clm->aiu->delete(clm->aiu);
    clm->aiu = NULL;
    delete(clm);
  
    return 0;
}


/*------------------------------------------------------------------*/
/**
 *  
 */
static int aiu_resend_callback(void *obj, unsigned int time){
    claimer *clm = obj;

    clm->last_interval = clm->last_interval * 2;  
    clm->resend_timer = clm->t->set(clm->t, clm->last_interval, aiu_resend_callback, clm);

    clm->aiu->send(clm->aiu);

    return 0;
}
/*------------------------------------------------------------------*/
static void claim(claimer *clm) {
    addr_range *ar;

    ar = clm->aap->arec->sel_addr(clm->aap->arec, clm->numallocs, clm->end_time);
    if(ar == NULL) {
        if(clm->cfg->scope == CFG_LARGE_SCOPE) {
            /* Create the ANA message. */
            aap_ana *ana_msg = aap_ana_new(clm->n, clm->ns, clm->t, clm->cfg,
                                           clm->aap, clm->numallocs,
                                           clm->end_time,
                                           clm->aap->next_rseq(clm->aap));
            /* Set a timer for retransmission. */
            ana_msg->t->set(ana_msg->t, ana_msg->resend_time,
                            ana_msg->resender, ana_msg);
            /* send ANA message */
            ana_msg->send(ana_msg);
        } else {
            /* log message */
        }
    } else {
        clm->active = 1;
        clm->alloc = allocation_new(0, ar, time(NULL), ALLOCATION_OWN);
        clm->aclm->set_allocation(clm->aclm, clm->alloc);
        clm->aclm->send(clm->aclm);
    
        /* set up the resend time and the claim timer */
    
        clm->last_interval = clm->cfg->RESEND_WAIT;
        clm->resend_timer = clm->t->set(clm->t, clm->cfg->RESEND_WAIT, resend_callback, clm);
    
        clm->claim_timer = clm->t->set(clm->t, clm->cfg->ANNOUNCE_WAIT, claim_callback, clm);

    }

}
/*------------------------------------------------------------------*/
static void cancel(claimer *clm) {
    /* cancel the timers */
    clm->t->unset(clm->t, clm->resend_timer);
    clm->t->unset(clm->t, clm->claim_timer);
    clm->active = 0;
}
/*------------------------------------------------------------------*/
static int retry_in_callback(void *obj, unsigned int time) {
    claimer *clm = obj;
    clm->alloc->delete(&(clm->alloc));
    clm->alloc = NULL;
    clm->claim(clm);
    return 0;
}

static void retry_in(claimer *clm, unsigned int seconds) {
    cancel(clm);
    clm->t->set(clm->t, seconds, retry_in_callback, clm);
}
/*------------------------------------------------------------------*/
static void retry(claimer *clm) {
    cancel(clm);  
    clm->alloc->delete(&(clm->alloc));
    clm->alloc = NULL;
    /* retry the claim. now the allocation record contains
       the already allocated addresses, so that arec->sel_addr will
       give a clean address range. */
    clm->claim(clm);
}
/*------------------------------------------------------------------*/
static int is_active(claimer *clm) {
    return clm->active;
}
/*------------------------------------------------------------------*/
static void delete(claimer *clm) {
    if(clm->aclm != NULL) {
        clm->aclm->delete(clm->aclm);
    }
    if(clm->aiu != NULL) {
        clm->aiu->delete(clm->aiu);
    }
    free(clm);
}
/*------------------------------------------------------------------*/
static addr_range *range(claimer *clm) {
    return clm->alloc->range(clm->alloc);
}
/*------------------------------------------------------------------*/
claimer *claimer_new(aap *aap, int numallocs, unsigned int end_time,
		     aap_callback_func f, void *obj) {
    claimer *clm = malloc(sizeof(claimer));

    clm->n = aap->net;
    clm->ns = aap->out_ns;
    clm->t = aap->tim;
    clm->cfg = aap->cfg;
    clm->aap = aap;

    clm->numallocs = numallocs;
    clm->end_time = end_time;

    clm->f = f;
    clm->obj = obj;
  
    clm->aclm = aap_aclm_new(clm->n, clm->ns, clm->t, clm->cfg, clm->aap,
                             NULL, clm->aap->next_rseq(clm->aap));

    clm->aiu = NULL;
    clm->claim = claim;
    clm->delete = delete;
    clm->retry = retry;
    clm->retry_in = retry_in;
    clm->cancel = cancel;
    clm->range = range;
    clm->is_active = is_active;
    return clm;
}
