/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains constructor and methods for an addr_rec object. 
 * The address record contains all the addresses allocated and all client 
 * allocations made for these addresses.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <memory.h>

#include "addr_rec.h"
#include "aap.h"
#include "madcap_alloc.h"

/* TODO: Free all temp */

typedef struct temp_addr_rec_ temp_addr_rec;
typedef struct temp_rem_allocs_ temp_rem_allocs;
typedef struct temp_rem_addr_ temp_rem_addr;
typedef struct temp_lease_id_ temp_lease_id;

typedef int (* matching_function)(void *obj, void *element);

static int select_allocs(addr_rec *a_rec, madcap_plist *temp_list,
                         madcap_alloc_flag flag,
                         int req_no_addr, int req_min_no_addr,
                         uint32 req_lease_time, 
                         uint32 req_min_lease_time,
                         uint32 *alloc_lease_time,
                         uint32 *alloc_start_time,
                         madcap_plist *alloc_addr_ranges);

static void check_addr_avail(addr_rec *a_rec, madcap_plist *address_list);
static void free_madcap_alloc(madcap_alloc *mc_alloc);
static int callback_aap_timer(void *obj, unsigned int dummy);
static void sort_client_allocs(madcap_plist *client_allocs);

static int has_called_aap = 0;
static tim_id_t tim_id = 0;

/* Temp object used for callback_allocated function */
struct temp_addr_rec_ {
    addr_rec *a_rec;
    uint32 scope_id;
};

/* Temp object used for callback_rem_allocs function */
struct temp_rem_allocs_ {
    madcap_plist *addr_list;
    uint32 end_time;
    int timeout;
};

/* Temp object used for callback_rem_addr function */
struct temp_rem_addr_ {
    addr_rec *a_rec;
    addr_range *a_range;
};

/* Temp object used for comparing lease_id */
struct temp_lease_id_ {
    uint8 *lease_id;
    uint16 lease_id_len;   
};

/* Free memory used for a madcap_alloc object */
static void free_madcap_alloc(madcap_alloc *mc_alloc){
    int i;
    madcap_plist *client_list = mc_alloc->client_allocs_list;
    for(i = 0;i < client_list->length(client_list);i++) {
        free(client_list->elem_at(client_list, i));
    }
    client_list->delete(client_list);
    free(mc_alloc);
}

/* Write address record to file */
static void sync_cache(addr_rec *a){
    char filename[50];
    char filename_bak[50];
    FILE *f;
    int i, j, k;

    if(a->cfg->get_addr_rec_file(a->cfg, filename) != 0) {
            fprintf(stderr, "addr_rec: Got no filename from cfg-file\n");
            exit(1);
    }    
    /*
      Rename the old cache file to be used as backup in case of a crash while
      creating the new cache file.
    */
    sprintf(filename_bak, "%s.BAK", filename);
    rename(filename, filename_bak);
    f = fopen(filename, "w");
    if(f == NULL) {
        fprintf(stderr, "Failed to open file for cache.\n");
        exit(EXIT_FAILURE);
    }

    DEBUG('a', "Writing to file.\n");
    
    fprintf(f, "addr_rec {\n");
    /* Write each address to file */
    for(i = 0;i < a->alloc_list->length(a->alloc_list);i++) {
        madcap_alloc *tmp_alloc =
            (madcap_alloc *)a->alloc_list->elem_at(a->alloc_list, i);
        fprintf(f, "\talloc(%lu, %lu, %lu){\n",
                tmp_alloc->address,
                tmp_alloc->scope_id,
                tmp_alloc->address_end_time);
        /* Write each client allocation for the address to file */
        for(j = 0;j < tmp_alloc->client_allocs_list->
                length(tmp_alloc->client_allocs_list);j++) {
            madcap_client_alloc *tmp_calloc = (madcap_client_alloc *)
                tmp_alloc->client_allocs_list->
                elem_at(tmp_alloc->client_allocs_list, j);
            fprintf(f, "\t\tcalloc(");
            fprintf(f, "%u", (unsigned char)(*tmp_calloc->lease_id));
            for(k = 1;k < tmp_calloc->lease_id_length;k++) {
                fprintf(f, ":%u",
                        (unsigned char)(*(tmp_calloc->lease_id + k)));
            }
            fprintf(f, ",%hu, %d, %lu, %lu);\n", tmp_calloc->lease_id_length,
                    tmp_calloc->flag, tmp_calloc->start_time,
                    tmp_calloc->end_time);
        }
        fprintf(f, "\t};\n");
    }
    fprintf(f, "};\n");
    fclose(f);
}


/* Function used for comparing two lease id's */
static int compare_lease_id(temp_lease_id *first_lid,
                            temp_lease_id *other_lid){
    int i;
    if(first_lid->lease_id_len != other_lid->lease_id_len) {
        return 0;
    }
    for(i = 0;i < first_lid->lease_id_len;i++) {
        if(first_lid->lease_id[i] != other_lid->lease_id[i]) {
            return 0;
        }
    }
    return 1;
}

/* "Matching function" for searching lease_id in client_allocs_list. The 
   function is to be used as an argument to the function "find_element" 
   in madcap_plist. */
static int search_lease_id(void *obj, void *element){

    madcap_client_alloc *tmp = (madcap_client_alloc *) element;
    temp_lease_id *first = NULL;
    temp_lease_id *other = (temp_lease_id *) malloc(sizeof(temp_lease_id));
    
    first = (temp_lease_id *) obj;   
    other->lease_id = tmp->lease_id;
    other->lease_id_len = tmp->lease_id_length;

    if(compare_lease_id(first, other) == 1) {
        free(other);
        return 1;
    } else { 
        free(other);
        return 0;
    }
}

/* "Matching function" for searching the end_time of an allocation in 
   client_allocs_list. The function is to be used as an argument to the 
   function "find_element" in madcap_plist. */
static int search_alloc_end_time(void *obj, void *element){

    uint32 *pointer = obj;
    uint32 first = *pointer;
    madcap_client_alloc *tmp = (madcap_client_alloc *) element;
    uint32 other = tmp->end_time;
    if(first == other) {
        return 1;
    } else {
        return 0;
    }
}

/* "Matching function" for searching the end_time of an address in 
   alloc_list. The function is to be used as an argument to the 
   function "find_element" in madcap_plist. */
static int search_address_end_time(void *obj, void *element){

    uint32 *pointer = obj;
    uint32 first = *pointer;
    madcap_alloc *tmp = (madcap_alloc *) element;
    uint32 other = tmp->address_end_time;
    if(first == other) {
        return 1;
    } else {
        return 0;
    }
}

/* Callback function used when client allocations has expired */
static int callback_rem_allocs(void *obj, unsigned int dummy){
    
    int i, j;
    madcap_plist *addr_list = ((temp_rem_allocs *)obj)->addr_list;
    madcap_plist *client_allocs;
    uint32 end_time = ((temp_rem_allocs *)obj)->end_time;
    uint32 *p = malloc(sizeof(uint32));
    *p = end_time;

    DEBUG('a', "Callback for callback_rem_allocs.\n");

    for(i = 0;i<addr_list->length(addr_list);i++) {
        client_allocs = ((madcap_alloc *)addr_list->
                         elem_at(addr_list, i))->client_allocs_list;
        for(j = 0;j<client_allocs->length(client_allocs);j++) {
            client_allocs->remove_element(client_allocs,
                                          search_alloc_end_time,
                                          p);
        }
    }
    return 0;
}

/* Callback function used when an address has expired */
static int callback_rem_addr(void *obj, unsigned int dummy){
    uint32 f, l, end_time;
    uint32 *p;

    temp_rem_addr *tmp = (temp_rem_addr *) obj;
    addr_rec *self = (addr_rec *) tmp->a_rec;
    addr_range *ar = (addr_range *) tmp->a_range;
    
    DEBUG('a', "Callback for callback_rem_addr.\n");

    /* Convert to host byte order */
    f = ntohl(ar->first_addr(ar));
    l = ntohl(ar->last_addr(ar));
    end_time = ar->end_time(ar);
    p = malloc(sizeof(uint32));
    *p = end_time;
    
    /* Remove expired addresses from alloc_list */
    self->alloc_list->remove_element(self->alloc_list,
                                     search_address_end_time, p);
    DEBUG('a', "Length of alloc_list after removal: %d\n",
          (self->alloc_list)->length(self->alloc_list));

    free(p);
    /* Dummy return value */
    return 0;
}

/* Callback function, allows new calls for aap->allocate() */
static int callback_aap_timer(void *obj, unsigned int dummy){
    has_called_aap = 0;
    tim_id = 0;
    if(((addr_rec *)obj)->alloc_list->length(((addr_rec *)obj)->
                                             alloc_list) != 0) {
        check_addr_avail((addr_rec *)obj, ((addr_rec *)obj)->alloc_list);
    } else {
        DEBUG('a', "addr_rec: aap allocation timed out\n");
    }
    return 0;
}

/* Callback function used when addresses are allocated from aap */
static void callback_allocated(void *obj, addr_range *ar){

    uint32 f, l, i, end_time;
    int arg_time;
    madcap_alloc *mc_alloc;
    addr_range *a_range;
    temp_addr_rec *tmp = (temp_addr_rec *) obj;
    temp_rem_addr *tmp_rem;
    addr_rec *self = (addr_rec *) tmp->a_rec;
    uint32 scope_id = (uint32) tmp->scope_id;

    DEBUG('a', "Callback for callback_allocated.\n");

    if(ar == NULL) {
        fprintf(stderr, "addr_rec: Could not allocate addresses from aap\n");
        has_called_aap = 0;
        if(tim_id != 0) {
            self->tim->unset(self->tim, tim_id);
        }
        tim_id = 0;
        exit(1);       
    }
    
    /* Convert to host byte order */
    f = ntohl(ar->first_addr(ar));
    l = ntohl(ar->last_addr(ar));
    end_time = ar->end_time(ar);

    a_range = addr_range_new(ar->first_addr(ar), ar->last_addr(ar), end_time);
    arg_time = end_time - time(NULL);

    /* Insert allocated addresses into alloc_list */
    for (i = f;i<(l + 1);i++) {        
        mc_alloc = madcap_alloc_new(i, scope_id, end_time);
        (self->alloc_list)->insert(self->alloc_list, end_time, mc_alloc);
    }

    has_called_aap = 0;
    if(tim_id != 0) {
        self->tim->unset(self->tim, tim_id);
    }
    tim_id = 0;
 
    /* Set timer, use callback function to remove addresses when
       timer expires */
    tmp_rem = malloc(sizeof(temp_rem_addr));
    tmp_rem->a_rec = self;
    tmp_rem->a_range = a_range;
    self->tim->set(self->tim, arg_time,
                   callback_rem_addr, tmp_rem);

    sync_cache(self);
    
    DEBUG('a', "Addresses allocated.\n");
    DEBUG('a', "Length of alloc_list: %d.\n",
                    ((self->alloc_list)->length(self->alloc_list)));
}

/* Check that min no of addresses for each scope are available */
static void check_addr_avail(addr_rec *a_rec, madcap_plist *addr_list){
    int i, j, k;
    int tot_no_addr = 0;
    int free_addr = 0;
    int new_addr = 0;
    uint32 sco_id;
    madcap_list *aap_list;
    aap *aap_p;
    madcap_alloc *mc_alloc;

    int no_of_scopes = 0;
    scope *scopes;
    scope *scp;
    scope *tmp_scp;
    temp_addr_rec *temp;  
    unsigned int req_time;
 
    aap_list = a_rec->get_aap_list(a_rec);

    for(i = 0;i<aap_list->length(aap_list);i++) {
        sco_id = ((aap_list_item *)aap_list->elem_at(aap_list, i))->scope_id;
        aap_p = (aap *) ((aap_list_item *) aap_list->
                         elem_at(aap_list, i))->aap;
        for(j = 0;j<addr_list->length(addr_list);j++) {
            mc_alloc = (madcap_alloc *)addr_list->elem_at(addr_list, j);
            if((mc_alloc->scope_id == sco_id) &&
               (mc_alloc->address_end_time >
                (time(NULL) + ADDR_REC_MIN_TIME))) {
                tot_no_addr++;
                if(mc_alloc->client_allocs_list->
                   length(mc_alloc->client_allocs_list) == 0) {
                    free_addr++;
                }
            }
        }        
        k = a_rec->cfg->get_number_of_scopes(a_rec->cfg, &no_of_scopes);
        if((k != SUCCESS) || (no_of_scopes == 0)) {
            fprintf(stderr, "addr_rec: Got no scopes from cfg-file\n");
            exit(1);
        }
        k = a_rec->cfg->get_scopes(a_rec->cfg, &scopes);
        if((k != SUCCESS) || (no_of_scopes == 0)) {
            fprintf(stderr, "addr_rec: Got no scopes from cfg-file\n");
            exit(1);
        }
        /* get scope */
        for(k = 0;k<no_of_scopes;k++) {
            tmp_scp = &scopes[k];
            if(tmp_scp->addresses.start == sco_id) {
                scp = &scopes[k];
            }
        }
        if(scp == NULL) {
            fprintf(stderr, "addr_rec: No scope matching scope_id\n");
            exit(1);
        }
        DEBUG('a', "free_addr = %d\n", free_addr);
        DEBUG('a', "tot_no_addr = %d\n", tot_no_addr);
        DEBUG('a', "min_free_addr = %d\n", scp->min_free_addr);
        
        if(((free_addr*100)/tot_no_addr) < scp->min_free_addr) {
            
            /* Put addr_rec and scope in a struct, which is passed on via
               allocate to the receiving callback function */
            temp = (temp_addr_rec *)malloc(sizeof(temp_addr_rec));
            temp->a_rec = a_rec;
            temp->scope_id = (sco_id);
            req_time = time(NULL) + ADDR_REC_REQ_TIME;
            new_addr = ((tot_no_addr * scp->min_free_addr) / 100);
            has_called_aap = 1;
            tim_id = a_rec->tim->set(a_rec->tim,
                                     (aap_p->cfg->ANNOUNCE_WAIT * 2),
                                     callback_aap_timer,
                                     a_rec);
            aap_p->allocate(aap_p, new_addr, req_time,
                            callback_allocated, temp);
        }
    }
    tot_no_addr = 0;
    free_addr = 0;
    new_addr = 0;
}

/* Function whichs sorts a list of client allocations by the
   allocation end time */
static void sort_client_allocs(madcap_plist *client_allocs){
    int i;
    madcap_plist *tmp = madcap_plist_new();
    /* Put all elements in the new list. */
    for(i = 0;i < client_allocs->length(client_allocs);i++) {
        tmp->insert(tmp,
                    ((madcap_client_alloc *)client_allocs->
                     elem_at(client_allocs, i))->end_time,
                    client_allocs->elem_at(client_allocs, i));
    }
    /* Remove all from list. */
    while(!client_allocs->is_empty(client_allocs)) {
        client_allocs->remove(client_allocs, 0);
    }
    /* Move everything back again. */
    for(i = 0;i < tmp->length(tmp);i++) {
        client_allocs->insert(client_allocs,
                              ((madcap_client_alloc *)tmp->
                               elem_at(tmp, i))->end_time,
                              tmp->elem_at(tmp, i));
    }
    tmp->delete(tmp);
}

/* Function which searches through the temp_list and selects which
   addresses to allocate */
static int select_allocs(addr_rec *a_rec, madcap_plist *temp_list, 
                         madcap_alloc_flag flag, 
                         int req_no_addr, int req_min_no_addr,
                         uint32 req_lease_time,
                         uint32 req_min_lease_time,
                         uint32 *alloc_lease_time,
                         uint32 *alloc_start_time,
                         madcap_plist *alloc_addr_ranges){
    int i, j;
    int c = 0;
    uint32 temp_start_time, start_time, end_time;
    int alloc_time = 0;
    madcap_plist *address_list = a_rec->alloc_list;
    madcap_client_alloc *client_alloc;
    madcap_plist *acc_list = madcap_plist_new();
    madcap_plist *client_allocs;
    unsigned long f_addr, l_addr, e_time;
    addr_range *temp_addr_range;
    tim_id_t tid;
   
    for(i = 0;i<temp_list->length(temp_list);i++) {
        /* First start time to try */
        temp_start_time = ((madcap_client_alloc *)temp_list->
                           elem_at(temp_list, i))->start_time;

        /* Search start time and end time for all other addresses */
        for(j = 0;j<temp_list->length(temp_list);j++) {
            if(((madcap_client_alloc *)temp_list->elem_at(temp_list, j))->
               start_time <= temp_start_time) {
                if((((madcap_client_alloc *)temp_list->
                     elem_at(temp_list, j))->
                    end_time - temp_start_time) >= req_min_lease_time) {
                    /* Address can be used, put in acc_list */
                    acc_list->insert(acc_list,
                                     ((madcap_client_alloc *)temp_list->
                                      elem_at(temp_list, j))->end_time,
                                     temp_list->elem_at(temp_list, j));
                    c++;  
                }
            }
        }
        /* Enough addresses allocated with simultanous lifetime, give
           them all the same start time and end time */
        if(c >= req_min_no_addr) {
            start_time = temp_start_time;
            if(c >= req_no_addr) {
                if(((((madcap_client_alloc *)acc_list->
                      elem_at(acc_list, (req_no_addr - 1)))->end_time) >
                   (start_time + req_lease_time)) && (req_lease_time != 0)) {
                    end_time = start_time + req_lease_time;
                } else {
                    end_time = ((madcap_client_alloc *)acc_list->
                                elem_at(acc_list, (req_no_addr - 1)))->
                        end_time;
                }                
                /* Remove addresses exceeding requested number */
                while((acc_list->length(acc_list)) > req_no_addr) {
                    acc_list->remove(acc_list,
                                     (acc_list->length(acc_list) - 1));
                }
            } else {
                if(((((madcap_client_alloc *)acc_list->
                      elem_at(acc_list, (c - 1)))->end_time) >
                    (start_time + req_lease_time)) && (req_lease_time != 0)) {
                    end_time = start_time + req_lease_time;
                } else {
                    end_time = ((madcap_client_alloc *)acc_list->
                                elem_at(acc_list,
                                        (acc_list->length(acc_list) - 1)))->
                        end_time;
                }
            }
            /* Change temp allocations to "real" allocations
               in address list */
            for(c = 0;c < acc_list->length(acc_list);c++) {
                client_alloc = (madcap_client_alloc *)acc_list->
                    elem_at(acc_list, c);
                client_alloc->flag = flag;
                client_alloc->start_time = start_time;
                client_alloc->end_time = end_time;

                /* Set timer */
                if(c == 0) {
                    temp_rem_allocs *tim_args =
                        malloc(sizeof(temp_rem_allocs));
                    ASSERT(tim_args != NULL, "Memory allocation failed.");
                    tim_args->addr_list = address_list;
                    tim_args->end_time = client_alloc->end_time;
                    if((a_rec->cfg->
                        get_extra_allocation_time(a_rec->cfg,
                                                  a_rec->cfg->name,
                                                  &alloc_time)) != 0) {
                        fprintf(stderr, "addr_rec: No configuration file\n");
                        exit(1);
                    }
                    tim_args->timeout =
                        client_alloc->end_time - time(NULL) + alloc_time;
                    
                    tid = a_rec->tim->set(a_rec->tim, tim_args->timeout,
                                          callback_rem_allocs, tim_args);
                }
                client_alloc->tim_id = tid;
                /* Insert allocation in list. */
                client_alloc->owner->client_allocs_list->
                    insert(client_alloc->owner->client_allocs_list,
                           client_alloc->end_time,
                           client_alloc);
                /* Create addr_range for addresses */
                e_time = end_time;
                f_addr = ((madcap_alloc *)(client_alloc->owner))->address;
                l_addr = f_addr;
                e_time = end_time;
                temp_addr_range = addr_range_new(f_addr,
                                                 l_addr,
                                                 e_time);
                /* Insert into "return-list" */
                alloc_addr_ranges->insert(alloc_addr_ranges,
                                          temp_addr_range->
                                          first_addr(temp_addr_range),
                                          temp_addr_range);
            }            
            /* Return values */ 
            *alloc_lease_time = end_time - start_time;
            *alloc_start_time = start_time;
            
            /* Sort list with client allocations with end time as priority */
            for(c = 0;c < address_list->length(address_list);c++) {
                client_allocs = ((madcap_alloc *)address_list->
                                 elem_at(address_list, c))->client_allocs_list;
                sort_client_allocs(client_allocs);
            }
            return 0;
        }
        c = 0;
    }
    return -1;
}

/**
 * Allocate multicast addresses.
 *
 * @param object The address record that makes the allocation.
 * @param flag Indicates what type of allocation that should be made.
 * @param lease_id The unique lease identifier provided by the client.
 * @param lease_id_len The length of lease_id in bytes.
 * @param scope_id The scope ID of the multicast scope that the client desires.
 * @param req_lease_time The lease time requested by the client. O indicates
 * that the maximum avilable time should be used.
 * @param req_min_lease_time The minimum acceptable lease time that the client
 * can accept. O indicates no minimum time.
 * @param req_start_time The start time requested by the client for a
 * preallocation. O indicates as soon as possible.
 * @param req_max_start_time The maximum acceptable start time that the client
 * can accept. O indicates no maximum time.
 * @param req_no_addr The number of addresses that the client desires.
 * @param req_min_no_addr The minimum number of addresses that the client can
 * accept.
 * @param req_addr_ranges The list of address ranges that the client has
 * requested. NULL indicates no special desire.
 * @param alloc_lease_time The lease time that the server has allocated.
 * @param alloc_start_time The start time that the server has allocated.
 * @param alloc_addr_ranges The list of address ranges that the server has
 * allocated.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 */
static int pub_alloc_addr(addr_rec *object, 
                          madcap_alloc_flag flag,
                          uint8 *lease_id,
                          uint16 lease_id_len,
                          uint32 scope_id,
                          uint32 req_lease_time,
                          uint32 req_min_lease_time,
                          uint32 req_start_time,
                          uint32 req_max_start_time,
                          uint16 req_no_addr,
                          uint16 req_min_no_addr,
                          madcap_plist *req_addr_ranges,
                          uint32 *alloc_lease_time,
                          uint32 *alloc_start_time,
                          madcap_plist *alloc_addr_ranges){

    int i, j, m;
    int alloc_time;
    madcap_plist *address_list = (madcap_plist *) object->alloc_list;
    madcap_plist *client_allocs = NULL;
    madcap_plist *temp_list;
    madcap_alloc *current_addr = NULL;
    
    /* These are used for searching client allocations */
    madcap_client_alloc *client_alloc = NULL;
    madcap_client_alloc *new_alloc = NULL;
    madcap_client_alloc *prev_alloc = NULL;

    uint32 start_time, end_time, temp_time, current_time;
    temp_list = madcap_plist_new();
    ASSERT(temp_list != NULL, "List creation failed.");

    /* DISCOVER message not supported */
    if(flag == MADCAP_ALLOC_PREALLOC) {
        return -1;
    }
    
    if(object->cfg->get_extra_allocation_time(object->cfg,
                                              object->cfg->name,
                                              &alloc_time) != 0) {
        fprintf(stderr, "addr_rec: No configuration file\n");
        exit(1);
    }
    
    /* Default values */
    if(req_start_time == 0) {
        req_start_time = time(NULL);
    }
    if(req_max_start_time == 0) {
        req_max_start_time = ADDR_REC_REQ_TIME;
    }
    if(req_no_addr == 0) {
        req_no_addr = 1;
        req_min_no_addr = 1;
    }
    current_time = time(NULL);

    /* TODO: Improve search algorithm for finding time between already
       made allocations when options req_start_time and
       req_max_start_time are used */

    for(i = 0;i<address_list->length(address_list);i++) {
        current_addr = address_list->elem_at(address_list, i);
        if(current_addr->scope_id == scope_id) { /* Within scope */
            client_allocs = current_addr->client_allocs_list;
            start_time = 0;
            end_time = 0;
            
            if(client_allocs->is_empty(client_allocs)) { /* Unused address */
                if(current_addr->address_end_time >
                   (req_start_time + req_min_lease_time + alloc_time)) {
                    
                    /* Create madcap_client_alloc and insert into temp_list */
                    new_alloc = malloc(sizeof(madcap_client_alloc));
                    ASSERT(new_alloc != NULL,
                           "Client allocation creation failed.");
                    new_alloc->lease_id =
                        (uint8 *)malloc(lease_id_len * sizeof(uint8));
                    memcpy(new_alloc->lease_id, lease_id, lease_id_len);
                    new_alloc->lease_id_length = lease_id_len;
                    new_alloc->flag = MADCAP_ALLOC_TEMP;
                    new_alloc->start_time = req_start_time;
                    new_alloc->end_time =
                        (current_addr->address_end_time - alloc_time);
                    new_alloc->owner = current_addr;
                    temp_list->insert(temp_list,
                                      new_alloc->end_time,
                                      new_alloc);
                }                
            } else {  /* Address with other client allocations */
                /* Search through all client allocations */
                for(j = 0;j < client_allocs->length(client_allocs);j++) {
                    client_alloc = client_allocs->elem_at(client_allocs, j);
                    start_time = 0;
                    end_time = 0;
                    
                    /* Handle first alloc separately since there is no
                       previous alloc */
                    if(j == 0) {
                        temp_time = 0;
                    } else {
                        temp_time = prev_alloc->end_time;
                    }
                    /* Last allocation, check against address end_time instead
                       of previous client alloc end_time */
                    if(j == (client_allocs->length(client_allocs) - 1)) {
                        if((client_alloc->start_time >
                            (req_start_time + req_min_lease_time +
                             alloc_time)) &&
                           ((temp_time + alloc_time) < req_start_time)) {
                            start_time = req_start_time;
                            end_time = (client_alloc->start_time - alloc_time);

                        } else if(((client_alloc->end_time + alloc_time) <
                                  req_start_time) && 
                                 (current_addr->address_end_time > 
                                  (req_start_time + req_min_lease_time +
                                   alloc_time))) {
                            start_time = req_start_time;
                            end_time = current_addr->address_end_time -
                                alloc_time;
                            
                        } else if((client_alloc->end_time + alloc_time <
                                  req_max_start_time) &&
                                 (current_addr->address_end_time >
                                  (req_max_start_time + req_min_lease_time +
                                   alloc_time))){
                            
                            start_time = (client_alloc->end_time + alloc_time);
                            end_time = (current_addr->address_end_time -
                                        alloc_time);
                        }
                    } else { /* This client_alloc is not the last */
                        
                        if((client_alloc->start_time >
                            (req_start_time + req_min_lease_time + alloc_time))
                           && ((temp_time + alloc_time) < req_start_time)) {
                            
                            start_time = req_start_time;
                            end_time = (client_alloc->start_time - alloc_time);
                        } else if((client_alloc->start_time >
                                  (req_max_start_time + req_min_lease_time +
                                   alloc_time)) &&
                                 ((temp_time + alloc_time) <
                                  req_max_start_time)) {
                            
                            start_time = (temp_time + alloc_time);
                            end_time = (client_alloc->start_time - alloc_time);
                        }
                    }
                    /* This address can be used */                    
                    if((start_time != 0) && (end_time != 0)) {
                        new_alloc = malloc(sizeof(madcap_client_alloc));
                        ASSERT(new_alloc != NULL,
                               "Client allocation creation failed.");
                        new_alloc->lease_id =
                            (uint8 *)malloc(lease_id_len * sizeof(uint8));
                        memcpy(new_alloc->lease_id, lease_id, lease_id_len);
                        new_alloc->lease_id_length = lease_id_len;
                        new_alloc->flag = MADCAP_ALLOC_TEMP;
                        new_alloc->start_time = start_time;
                        new_alloc->end_time = end_time;
                        new_alloc->owner = current_addr;
                        temp_list->insert(temp_list,
                                          end_time,
                                          new_alloc);
                    }
                    prev_alloc = client_alloc;
                }
            }
        }
    }
    DEBUG('a', "addr_rec: Lenght of temp_list = %d.\n",
          temp_list->length(temp_list));
    m = select_allocs(object, temp_list, flag, req_no_addr, req_min_no_addr,
                      req_lease_time, req_min_lease_time, alloc_lease_time,
                      alloc_start_time, alloc_addr_ranges);
    temp_list->delete(temp_list);
    if(m == 0) {
        sync_cache(object);
        if(has_called_aap == 0) {
            check_addr_avail(object, address_list);
        }
        return SUCCESS;
    } else {
        if(address_list->length(address_list) < req_no_addr) {
            mc_error->notify(mc_error, MC_ERR_OUTOFADDR,
                             "Not enough addresses available.\n");
        } else {
            mc_error->notify(mc_error, MC_ERR_ADDRLTIME,
                             "Not enough lifetime on available addresses.\n");
        }
        return FAILURE;  
    }
}

/**
 * This function is not implemented!
 *
 * Renew multicast addresses. This function is used for changing the 
 * lease time of an allocation already made.
 *
 * @param object The address record that makes the allocation.
 * @param lease_id The lease identifier for the address to be renewed.
 * @param lease_id_len The length of lease_id in bytes.
 * @param req_lease_time The new lease time requested by the client. 
 * O indicates that the maximum avilable time should be used.
 * @param req_min_lease_time The minimum acceptable lease time that the 
 * client can accept. O indicates no minimum time.
 * @param req_start_time The new start time requested by the client. 
 * O indicates as soon as possible.
 * @param req_max_start_time The maximum acceptable start time that the 
 * client can accept. O indicates no maximum time.
 * @param alloc_scope_id The scope of the renewed addresses.
 * @param alloc_lease_time The lease time that the server has allocated.
 * @param alloc_start_time The start time that the server has allocated.
 * @param alloc_addr_ranges The list of address ranges that the server has
 * renewed.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 */
static int pub_renew_addr(addr_rec *object, 
                          uint8 *lease_id,
                          uint16 lease_id_len,
                          uint32 req_lease_time,
                          uint32 req_min_lease_time,
                          uint32 req_start_time,
                          uint32 req_max_start_time,
                          uint32 *alloc_scope_id,
                          uint32 *alloc_lease_time,
                          uint32 *alloc_start_time,
                          madcap_plist *alloc_addr_ranges) {


    /* TODO: Not implemented yet!
     * Check that an allocation with this lease_id already exists. 
     * If the lease time of the allocation can be renewed, change 
     * the start_time and the end_time of the allocation and unset 
     * the timer for the old end_time and set the timer for the new 
     * end_time.     
     */        

    /* FIXME: Code only for testing purposes */
    /* Remove! */
    /* Now we set the new 'lease_time' to  */
    /* 'requested_lease_time' */
    address_range *ar = (address_range *)addr_range_new(0x00000001,
                                                        0x00000003,
                                                        4000);

    DEBUG('a', "addr_rec: renew_address not supported yet.\n");
    *alloc_lease_time = req_lease_time;
    *alloc_start_time = 0;
    *alloc_scope_id   = 0xFFFFFFFF;
    alloc_addr_ranges->insert(alloc_addr_ranges, 0, ar);
    
    return 0;
}

/**
 * This function is not fully implemented!
 *
 * Confirm an already made pre-allocation.
 *
 * @param object The address record that made the pre-allocation.
 * @param lease_identifier The lease identifier for the pre-allocation to be 
 * confirmed.
 * @param lease_identifier_length The length of lease_id in bytes.
 * @param conf_pre The list of address ranges that the server has
 * allocated.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 */
static int pub_confirm_prealloc(addr_rec *object, 
                                uint8 *lease_identifier, 
                                uint16 lease_identifier_length,
                                madcap_plist *conf_pre){
    int i, j;
    madcap_alloc *mc_alloc;
    madcap_plist *addr_list;
    madcap_plist *client_list;
    temp_lease_id *first_lid;
    temp_lease_id *other_lid;
    madcap_client_alloc *client_all;

    addr_list = object->alloc_list;
    first_lid->lease_id = lease_identifier;
    first_lid->lease_id_len = lease_identifier_length;

    /* TODO: Not fully implemented yet!
     * Check that all arguments are the same as they were in the original 
     * DISCOVER message.
     * Possible errors:
     * MC_ERR_LEASE_ID, Unknown lease ID. 
     * MC_ERR_LEASE, The addresses were not the ones that had been allocated.
     */
    for(i = 0;i < addr_list->length(addr_list);i++) {
        mc_alloc = (madcap_alloc *)addr_list->elem_at(addr_list, i);
        client_list = mc_alloc->client_allocs_list;
        for(j = 0;j < client_list->length(client_list);j++) {
            client_all = (madcap_client_alloc *)client_list->
                elem_at(client_list, j);
            other_lid->lease_id = client_all->lease_id;
            other_lid->lease_id_len = client_all->lease_id_length;
            if((client_all->flag == MADCAP_ALLOC_PREALLOC) &&
                (compare_lease_id(first_lid, other_lid) == 1)) {
                client_all->flag = MADCAP_ALLOC_ALLOC;
            }
        }
    }
    return -1;
}

/**
 * This function is not fully implemented!
 *
 * Release an already made pre-allocation.
 *
 * @param object The address record that made the pre-allocation.
 * @param lease_identifier The lease identifier for the pre-allocation to be 
 * released.
 * @param lease_identifier_length The length of lease_id in bytes.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 */
static int pub_release_prealloc(addr_rec *object, 
                                uint8 *lease_identifier,
                                uint16 lease_identifier_length){
    int i, j;
    tim_id_t tid;
    madcap_plist *addr_list = object->alloc_list;
    madcap_plist *client_list;
    temp_lease_id *lid = malloc(sizeof(temp_lease_id));
    ASSERT(lid != NULL, "Memory allocation failed.");
    lid->lease_id = lease_identifier;
    lid->lease_id_len = lease_identifier_length;

    /* TODO: Not fully implemented yet!
     * Check that all arguments are the same as they were in the original 
     * DISCOVER message.
     * Possible errors:
     * MC_ERR_LEASE_ID, Unknown lease ID. 
     * MC_ERR_LEASE, The addresses were not the ones that had been allocated.
     */
    for(i = 0;i < addr_list->length(addr_list);i++) {
        client_list = ((madcap_alloc *)addr_list->elem_at(addr_list, i))->
            client_allocs_list;
        tid = ((madcap_client_alloc *)client_list->
               find_element(client_list, search_lease_id, lid))->tim_id;
        client_list->remove_element(client_list, search_lease_id, lid);
    }
    j = ((tim *)object->tim)->unset(((tim *)object->tim), tid);
    return -1;
}

/**
 * Release an already made allocation.
 *
 * @param object The address record that made the allocation.
 * @param lease_identifier The lease identifier for the allocation to be 
 * released.
 * @param lease_identifier_length The length of lease_id in bytes.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 */
static int pub_release_alloc(addr_rec *object, 
                             uint8 *lease_identifier, 
                             uint16 lease_identifier_length){
    int i = 0;
    tim_id_t tid = 0;
    int result = FAILURE;
    madcap_plist *addr_list = object->alloc_list;
    madcap_plist *client_list = NULL;
    madcap_client_alloc *temp = NULL;
    temp_lease_id *lid = (temp_lease_id *) malloc(sizeof(temp_lease_id));
    lid->lease_id = lease_identifier;
    lid->lease_id_len = lease_identifier_length;

    for(i = 0;i < addr_list->length(addr_list);i++) {
        client_list = ((madcap_alloc *)addr_list->elem_at(addr_list, i))->
            client_allocs_list;
        ASSERT(client_list != NULL,
               "addr_rec:pub_release_alloc: client_list == NULL");
        
        temp = (madcap_client_alloc *)
            (client_list->find_element(client_list, search_lease_id, lid));
        if(temp != NULL) {
            tid = temp->tim_id;
            client_list->remove_element(client_list, search_lease_id, lid);
            DEBUG('a', "Releasing allocation.\n");
            result = SUCCESS;
        }
    }

    if(result != SUCCESS) {
        mc_error->notify(mc_error, MC_ERR_LEASE_ID,
                         "addr_rec:pub_release_alloc: Unknown lease id");
    }
    if(tid != 0) {
        ((tim *)object->tim)->unset(((tim *)object->tim), tid);
    }
    sync_cache(object);
    free(lid);
    return result;
}

/**
 * Get the aap list.
 *
 * @param object The address record which is holding the aap list.
 * @return A pointer to the aap list.
 */
static madcap_list *pub_get_aap_list(addr_rec *object){
    return object->aap_list;
}

/**
 * The constructor of an addr_rec object.
 *
 * @param cfg The current madcap configuration.
 * @param tim The timer.
 * @param aap_list The list containing the aap:s.
 * @param cache_list The list containing the allocations read from file.
 * @return A pointer to an addr_rec object.
 */
addr_rec *addr_rec_new(madcap_cfg *cfg, tim *tim, madcap_list *aap_list,
                       madcap_list *cache_list){
    
    temp_addr_rec *temp;
    aap *aap_p;
    int i, j;
    int no_of_scopes = 0;
    unsigned int req_time;
    scope *scopes;
    scope *scp;
    scope *tmp_scp;
    madcap_alloc *mc_alloc;
    madcap_plist *client_allocs;
    madcap_client_alloc *single_alloc;
    temp_rem_addr *tmp_rem;
    addr_range *ar;
    temp_rem_allocs *tim_args;
    tim_id_t tid;
    tim_id_t *tmp_tid;
    madcap_plist *client_time_list;
    madcap_plist *addr_time_list;
    
    addr_rec *addr_rec = malloc(sizeof(struct addr_rec_));
    ASSERT(addr_rec != NULL, "Memory allocation failed.");
    req_time = time(NULL) + ADDR_REC_REQ_TIME;
    addr_rec->cfg = cfg;
    addr_rec->tim = tim;
    addr_rec->aap_list = aap_list;
    addr_rec->alloc_list = madcap_plist_new();
    ASSERT(addr_rec->alloc_list != NULL, "Priority List creation failed.");
    client_time_list = madcap_plist_new();
    ASSERT(client_time_list != NULL, "Priority List creation failed.");
    addr_time_list = madcap_plist_new();
    ASSERT(addr_time_list != NULL, "Priority List creation failed.");

    addr_rec->alloc_addr = pub_alloc_addr;
    addr_rec->renew_addr = pub_renew_addr;
    addr_rec->confirm_prealloc = pub_confirm_prealloc;
    addr_rec->release_prealloc = pub_release_prealloc;
    addr_rec->release_alloc = pub_release_alloc;
    addr_rec->get_aap_list = pub_get_aap_list;

    if(cfg == NULL) {
        fprintf(stderr, "addr_rec: No configuration file\n");
        exit(1);
    }
    j = cfg->get_number_of_scopes(cfg, &no_of_scopes);
    if((j != SUCCESS) || (no_of_scopes == 0)) {
        fprintf(stderr, "addr_rec: Got no scopes from cfg-file\n");
        exit(1);
    }
    j = cfg->get_scopes(cfg, &scopes);
    if((j != SUCCESS) || (no_of_scopes == 0)) {
        fprintf(stderr, "addr_rec: Got no scopes from cfg-file\n");
        exit(1);
    }

    /* First server start-up */
    if(cache_list == NULL) {

        /* For each aap in the list: */
        for (i = 0;i < aap_list->length(aap_list);i++) {
            
            /* Get aap from aap_list */
            aap_p = (aap *) ((aap_list_item *) aap_list->
                             elem_at(aap_list, i))->aap;
            
            /* Put addr_rec and scope in a struct, which is passed on via
               allocate to the receiving callback function */
            temp = (temp_addr_rec *)malloc(sizeof(temp_addr_rec));
            temp->a_rec = addr_rec;
            temp->scope_id = (((aap_list_item *)
                               aap_list->elem_at(aap_list, i))->scope_id);
            /* get scope */
            for(j = 0;j < no_of_scopes;j++) {
                tmp_scp = &scopes[j];
                if(tmp_scp->addresses.start == temp->scope_id) {
                    scp = &scopes[j];
                }
            }
            if(scp == NULL) {
                fprintf(stderr, "addr_rec: No scope matching scope_id\n");
                exit(1);
            }
            /* Allocate addresses from aap */
            has_called_aap = 1;
            tim_id = tim->set(tim, (aap_p->cfg->ANNOUNCE_WAIT * 2),
                              callback_aap_timer, addr_rec);
            aap_p->allocate(aap_p, scp->start_no_addr, req_time,
                            callback_allocated, temp);
        }
    } else {
        /* Put all structs from cache_list into addr_rec */
        for(i = 0;i < cache_list->length(cache_list);i++) {
            addr_rec->alloc_list->insert(addr_rec->alloc_list, i,
                                         cache_list->elem_at(cache_list, i));
        }
        for(i = 0;i < addr_rec->alloc_list->length(addr_rec->alloc_list);i++) {
            mc_alloc = addr_rec->alloc_list->elem_at(addr_rec->alloc_list, i);
            client_allocs = mc_alloc->client_allocs_list;
            for(j = 0;j < client_allocs->length(client_allocs);j++) {
                single_alloc = client_allocs->elem_at(client_allocs, j);
                /* Remove expired client allocations */
                if(single_alloc->end_time < time(NULL)) {
                    free(single_alloc);
                } else {
                    /* Set timer, use callback function to remove allocations
                       when timer expires */
                    if((tmp_tid = client_time_list->
                        find_elem_with_prio(client_time_list, single_alloc->
                                            end_time)) != NULL) {
                        single_alloc->tim_id = *tmp_tid;
                    } else {
                        tim_args = malloc(sizeof(temp_rem_allocs));
                        ASSERT(tim_args != NULL, "Memory allocation failed.");
                        tim_args->addr_list = addr_rec->alloc_list;
                        tim_args->end_time = single_alloc->end_time;
                        tim_args->timeout = single_alloc->
                            end_time - time(NULL);
                        tid = addr_rec->tim->set(addr_rec->tim,
                                                 tim_args->timeout,
                                                 callback_rem_allocs,
                                                 tim_args);
                        single_alloc->tim_id = tid;
                        client_time_list->insert(client_time_list,
                                                 single_alloc->end_time,
                                                 &tid);
                    }
                }
            }
            if(mc_alloc->address_end_time < time(NULL)) {
                /* Remove expired addresses */
                free_madcap_alloc(mc_alloc);
            } else {
                /* Set timer, use callback function to remove addresses when
                   timer expires */
                if((tmp_tid = addr_time_list->
                    find_elem_with_prio(addr_time_list, mc_alloc->
                                        address_end_time)) != NULL) {
                    /* Timer already set for this address_end_time */
                } else {
                    ar = addr_range_new(0, 0, mc_alloc->address_end_time);
                    tmp_rem = malloc(sizeof(temp_rem_addr));
                    tmp_rem->a_rec = addr_rec;
                    tmp_rem->a_range = ar;
                    addr_rec->tim->set(addr_rec->tim,
                                       (mc_alloc->
                                        address_end_time - time(NULL)),
                                       callback_rem_addr, tmp_rem);
                    addr_time_list->insert(addr_time_list,
                                           mc_alloc->address_end_time,
                                           mc_alloc);
                }
            }
        }
        sync_cache(addr_rec);
        check_addr_avail(addr_rec, addr_rec->alloc_list);
    }
    DEBUG('a', "Leaving constructor for addr_rec.\n");
    return addr_rec;
}
