/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains the different configuration settings for madcap.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include "madcap_list.h"
#include "madcap_error.h"
#include "madcap_cfg.h"
#include "madcap_plist.h"
#include "madcap_alloc.h"
#include "madcap_cfgparser.h"

/* this is the file that Bison parser reads */
extern FILE *mcin;

/* This is the list in which the Bison parser stores its output */
extern madcap_list *mc_cfgparser_list;

/**
 * Public method to get the name of the file where the config-values are stored.
 *
 * @param object The madcap_cfg object
 * @param filename The name of the file
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_addr_rec_file(madcap_cfg *object, char *filename){
    strcpy(filename, object->filename);
    return SUCCESS;
}

/**
 * Public method to get the list of scopes for the given object.
 *
 * @param object The madcap_cfg object
 * @param scps An array with scopes
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_scopes(madcap_cfg *object, scope *scps[]){
    *scps = object->scopes;
    return SUCCESS;
}

/**
 * Public method to get the port of the configuration with the given name.
 *
 * @param object The madcap_cfg object
 * @param name The name of the madcap_cfg object
 * @param port The port number
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_port(madcap_cfg *object, char *name, int *port){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *port = object->port;
        return SUCCESS;
    }
    mc_error->notify(mc_error, MC_ERR_CONF, "Could not get port");
    return FAILURE;
}

/**
 * Public method to get the clock skew allowance of the configuration.
 *
 * @param object The madcap_cfg object
 * @param name The name of the madcap_cfg object
 * @param csa The clock skew allowance
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_clock_skew_allowance(madcap_cfg *object, char *name,
                                        uint32 *csa){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *csa = object->CSA;
        return SUCCESS;
    }
    mc_error->notify(mc_error, MC_ERR_CONF,
                     "Could not get clock skew allowance");
    return FAILURE;
}

/**
 * Public method to get the delay of the configuration.
 *
 * @param object The madcap_cfg object.
 * @param name The name of the madcap_cfg object
 * @param delay The delay
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_discover_delay(madcap_cfg *object, char *name, int *delay){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *delay = object->DELAY;
        return SUCCESS;
    }
    mc_error->notify(mc_error, MC_ERR_CONF, "Could not get discover delay");
    return FAILURE;
}

/**
 * Public method to get the extra allocation time of the configuration.
 *
 * @param object The madcap_cfg object
 * @param name The name of the madcap_cfg object
 * @param alloc_time The extra allocation time
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_extra_allocation_time(madcap_cfg *object, char *name,
                                         int *alloc_time){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *alloc_time = object->ALLOC_TIME;
        return SUCCESS;
    }
     mc_error->notify(mc_error, MC_ERR_CONF,
                      "Could not get extra allocation time");
    return FAILURE;
}

/**
 * Public method to get the response delay of the configuration.
 *
 * @param object The madcap_cfg object
 * @param name The name of the madcap_cfg object
 * @param response_delay The response delay
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_no_response_delay(madcap_cfg *object, char *name,
                                    int *response_delay){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *response_delay = object->RESPONSE_DELAY;
        return SUCCESS;
    }
    mc_error->notify(mc_error, MC_ERR_CONF, "Could not get response delay");
    return FAILURE;
}

/**
 * Public method to get the offer hold of the configuration.
 *
 * @param object The madcap_cfg object
 * @param name The name of the madcap_cfg object
 * @param offer_hold The offer_hold should be >60 sec
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_offer_hold(madcap_cfg *object, char *name, int *offer_hold){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *offer_hold = object->OFFER_HOLD;
        return SUCCESS;
    }
    mc_error->notify(mc_error, MC_ERR_CONF, "Could not get offer hold");
    return FAILURE;
}

/**
 * Public method to get the response cache interval of the configuration.
 *
 * @param object The madcap_cfg object
 * @param name The name of the madcap_cfg object
 * @param rci The response cache interval should be 60<rci<300 seconds
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_response_cache_interval(madcap_cfg *object, char *name,
                                           int *rci){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *rci = object->RCI;
        return SUCCESS;
    }
    mc_error->notify(mc_error, MC_ERR_CONF,
                     "Could not get response cache interval");
    return FAILURE;
}

/**
 * Public method to get the xid reuse interval of the configuration.
 *
 * @param object The madcap_cfg object
 * @param name The name of the madcap_cfg object
 * @param xri The xid reuse interval should be 600 seconds
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_xid_reuse_interval(madcap_cfg *object, char *name,
                                      int *xri){
    if(name == NULL) {
        name = "default";
    }
    if(!(strcmp(object->name, name))) {
        *xri = object->XRI;
        return SUCCESS;
    }
    mc_error->notify(mc_error, MC_ERR_CONF,
                     "Could not get xid reuse interval");
    return FAILURE;
}

/**
 * Public method to get the number of scopes of the configuration.
 *
 * @param object The madcap_cfg-object.
 * @param count The number of scopes of the configuration.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_number_of_scopes(madcap_cfg *object, int *count){
    *count = object->scope_count;
    return SUCCESS;
}

/**
 * The mc_cfg_parse funciton returns the list with configure objects.
 *
 * @param filename The file to read from
 * @return A pointer to the list with configure objects
 */
madcap_list *mc_cfg_parse(char *filename){
    FILE *cfg_file;
    if(strcmp(filename, "default") == 0) {
        DEBUG('i', "MADCAP: Using default configuration\n"); 
    } else {
        cfg_file = fopen(filename, "r");
        mcin = cfg_file;
        if(mcparse() == 1) {
            fprintf(stderr, "Unrecoverable error in configuration file %s\n",
                    filename);
            exit(1);
        }
        return mc_cfgparser_list;
    }
    return NULL;
}

/**
 * The constructor of a madcap_cfg_new object.
 *
 * @return A madcap_cfg-object on success, a NULL-pointer in case
 * of failure.
 */
madcap_cfg *madcap_cfg_new(){
    madcap_cfg *object;
    name_set_node *ns_node;

    object = (madcap_cfg *)calloc(1, sizeof(madcap_cfg));
    ns_node = (name_set_node *)malloc(sizeof(name_set_node));
    
    /* Return NULL if the memory needed could not be allocated. */
    if((object == NULL) || (ns_node == NULL)) {
        fprintf(stderr,
                "Could not allocate enough memory for structs\n");
        return NULL;
    }
    
    /* Set the function-pointers */
    object->get_addr_rec_file = pub_get_addr_rec_file;
    object->get_scopes = pub_get_scopes;
    object->get_port = pub_get_port;
    object->get_clock_skew_allowance = pub_get_clock_skew_allowance;
    object->get_discover_delay = pub_get_discover_delay;
    object->get_extra_allocation_time = pub_get_extra_allocation_time;
    object->get_no_response_delay = pub_get_no_response_delay;
    object->get_offer_hold = pub_get_offer_hold; 
    object->get_response_cache_interval = pub_get_response_cache_interval;
    object->get_xid_reuse_interval = pub_get_xid_reuse_interval;
    object->get_number_of_scopes = pub_get_number_of_scopes;
    
    /* allocate memory for name and filename*/
    object->name = (char *)malloc(20 * sizeof(char));
    object->filename = (char *)malloc(50 * sizeof(char));
    
    /* Return NULL if the memory needed could not be allocated. */
    if(object->name == NULL || object->filename == NULL) {
        fprintf(stderr,
                "Could not allocate enough memory for structs\n");
        return NULL;
    }
   
    /* SET DEFAULT VALUES */
    strcpy(object->name, "default");
    object->type = CFG_MADCAP;
    strcpy(object->filename, "/tmp/madcap.cache");
    
    /* Create default scope */
    /* Set the address range for the scope */
    /* FIXME: set the right default values for the addresses */
    /* Setting localscope by now */
    /* We store the scope addresses in host byte order, so for the code below, */
    /* we use ntohl since inet_addr returns the 32 bit address in network */
    /* byte order. */
    (object->scopes)[0].addresses.start = ntohl(inet_addr("224.0.0.0"));
    (object->scopes)[0].addresses.end = ntohl(inet_addr("224.255.255.255"));
    /* Setting only one scope for now... */
    object->scope_count = 1;

    /* Set the name set for the scope */
    /* create the list */
    (object->scopes)[0].name_set = madcap_list_new();
    /* set the name set node */
    ns_node->language_tag = "en";
    ns_node->name = "Default scope";
    /* put the node in the list */
    (object->scopes)[0].name_set->insert(object->scopes->name_set,
                                          (void *)ns_node);
    /* Set the default ttl for the scope */
    (object->scopes)[0].ttl = 16;
    (object->scopes)[0].min_free_addr = 10;
    (object->scopes)[0].start_no_addr = 30;
    /* Finished default scope */
    object->port = 2535; /* port of selected scope */
    object->CSA = 1800; /* 30 minutes clock-skew-allowance */
    object->DELAY = 4; /*init=4s, discover delay, current retransmit timer*/
    object->ALLOC_TIME = 3600; /* 1 hour extra allocation time */
    object->RESPONSE_DELAY = 60; /* 60 s, no response delay */
    object->OFFER_HOLD = 70; /* 70s, offer hold >60 sec */
    object->RCI = 60; /* 60s, 60s<response cache interval<300s */
    object->XRI = 600; /* 10 minutes xid reuse interval */
    
    return object; 
}
