/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains functions that is used when the server receives a
 * RELEASE message. It checks the required and optional options.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */


#include "madcap_msg_release.h"
static int pub_receive(madcap_msg_release *object);
static int pub_delete(madcap_msg_release *object);
extern madcap_error *mc_error;

/**
 * Creates a new RELEASE message.
 * 
 * @param msg An incoming madcap_msg message with madcap_msg_type = MC_RELEASE.
 * @param addr The Madcap address record.
 * @param n A network abstraction object.
 * @param ns An abstract network connection.
 * @param mc The madcap object.
 * @return The newly created message. The message should be deleted using the
 * delete function. A return value of NULL indicates a failure to create the
 * new message.
 */
madcap_msg_release *madcap_msg_release_new(madcap_msg *msg,
                                           addr_rec *addr,
                                           net *n,
                                           net_socket *ns,
                                           madcap *mc) {

    madcap_msg_release *object;

    DEBUG('m',"Creating RELEASE message.\n");
    
    object = (madcap_msg_release *) malloc(sizeof(madcap_msg_release));
    ASSERT(object != NULL, "Memory allocation failed.");
    
    object->msg     = msg;
    msg->get_options_flag(msg, &(object->opt_types_incl));
    object->addr    = addr;
    object->n       = n;
    object->ns      = ns;
    object->mc      = mc;
    msg->get_xid(msg, &(object->xid));
    object->receive = pub_receive;
    object->delete  = pub_delete;

    return object;    
}


/*
 * Receives a RELEASE message.
 * The function takes the appropriate actions when a RELEASE message is
 * received. The function will send either an ACK or an NAK to the client as
 * response.
 *
 * @param object The object that is received.
 * @return 0 on success and -1 on failure.
 */
static int pub_receive(madcap_msg_release *object) {

    madcap_msg_ack *ack;
    madcap_msg_nak *nak;
    madcap_msg_format_options *format = madcap_msg_format_options_new();
    int send_ack                   = TRUE;
    int result                     = SUCCESS;
    int ok;                         /* Temp variable. */
    int check;                      /* Temp variable. */
    uint8 *opt_data;                /* Option data from incoming messages. */
    uint16 opt_data_len;            /* Length of opt_data. */

    uint8 *lease_id                = NULL; /* Used with option 3. */
    uint16 lease_id_len            = 0;    /* Used with option 3. */
    madcap_nak_error_code err_code = 0;    /* Used with option 16. */
    uint16 err_option              = 0;    /* Used with option 16. */
    int err_option_count           = 0;    /* Used with option 16. */
    int err_data_len               = 2;    /* Used with option 16. */
    madcap_error_type err_num      = 0;    /* Used with option 16. */

    DEBUG('m',"Receiving RELEASE message.\n");

    /* Check that required option LEASE ID (3) is present and valid. */
    ok = (object->opt_types_incl & 0x0004) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_LEASE_ID,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len > 0) && (opt_data != NULL)) {
            lease_id     = opt_data;
            lease_id_len = opt_data_len;
        } else {
            ok           = FALSE;
            DEBUG('m',"Lease ID present but not valid.\n");
        }
    }
    if(!ok) {
        err_code         = MC_NAK_INVALID_REQ;
        err_option       = 3;
        err_option_count++;
        send_ack         = FALSE;
        DEBUG('m',"Lease ID not present.\n");
    }
    
    /* Check if optional option FEATURE LIST (12) is present and valid. */
    /* TODO: */

    /*
     * Check that no forbiden options are present.
     * madcap_opt_type 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 13, 14, 15, 16 are not
     * allowed.
     */
    if((object->opt_types_incl & 0xf7fb) != 0) {
        err_code       = MC_NAK_INVALID_REQ;
        send_ack       = FALSE;
        if((object->opt_types_incl & 0x0001) != 0) {
            err_option = 1;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0002) != 0) {
            err_option = 2;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0008) != 0) {
            err_option = 4;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0010) != 0) {
            err_option = 5;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0020) != 0) {
            err_option = 6;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0040) != 0) {
            err_option = 7;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0080) != 0) {
            err_option = 8;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0100) != 0) {
            err_option = 9;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0200) != 0) {
            err_option = 10;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0400) != 0) {
            err_option = 11;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x1000) != 0) {
            err_option = 13;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x2000) != 0) {
            err_option = 14;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x4000) != 0) {
            err_option = 15;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x8000) != 0) {
            err_option = 16;
            err_option_count++;
        }
        DEBUG('m',"Forbiden option present.\n");
    }

    /* Process the release. */
    if(send_ack == TRUE) {
        ok =  object->addr->release_alloc(object->addr,
                                          lease_id,
                                          lease_id_len);

        if(ok == FAILURE) {
            send_ack       = FALSE;
            mc_error->madcap_errnum_collect(mc_error, &err_num);
            switch(err_num) {
            case MC_ERR_LEASE_ID:
                err_code   = MC_NAK_LEASE_ID_NOT_RECON;
                break;
            default:
                err_code   = MC_NAK_VALID_REQ_NOT_COMPLETED;
                err_option = 0xffff;
                err_option_count++;
            }
        }
    }

    /* Create ACK or NAK message. */
    object->data     = NULL;
    object->data_len = 0;
    ok               = SUCCESS;

    /* Add the SERVER IDENTIFIER (2) option (ACK and NAK). */
    ok = ok + format->server_id(format,
                                &(object->data),
                                &(object->data_len));

    /* Add the LEASE IDENTIFIER (3) option (ACK and NAK). */
    ok = ok + format->lease_id(format,
                               lease_id,
                               lease_id_len,
                               &(object->data),
                               &(object->data_len));

    /* Add the FEATURE LIST (12) option (ACK and NAK). */
    /* TODO: */

    /* Add the ERROR (16) option (NAK). */
    if(send_ack == FALSE) {
        if((err_code == MC_NAK_INVALID_REQ) ||
           (err_code == MC_NAK_VALID_REQ_NOT_COMPLETED)) {
            if(err_option_count != 1) {
                /* Problem was not with a specific option. */
                err_option = 0xffff;
            }
            err_option = htons(err_option);
            ok = ok + format->error(format,
                                    err_code,
                                    (uint8 *) &err_option,
                                    err_data_len,
                                    &(object->data),
                                    &(object->data_len));
        } else if(err_code == MC_NAK_LEASE_ID_NOT_RECON) {
            ok = ok + format->error(format,
                                    err_code,
                                    (uint8 *) NULL,
                                    0,
                                    &(object->data),
                                    &(object->data_len));
        } else {
            ok = ok + format->error(format,
                                    err_code,
                                    (uint8 *) NULL,
                                    err_data_len,
                                    &(object->data),
                                    &(object->data_len));
        }
    }
    
    /* Add the END (0) option (ACK and NAK). */
    ok = ok + format->end(format,
                          &(object->data),
                          &(object->data_len));

    /* Send message if options were added correctly. */
    if((ok == SUCCESS) && (send_ack == TRUE)) {
        /* Send ACK. */
        ack = madcap_msg_ack_new(object->n,
                                 object->ns,
                                 object->xid,
                                 object->data,
                                 object->data_len);
        if(ack != NULL) {
            ok = ack->send(ack);
            if(ok == SUCCESS) {
                /* Save message so it can be resent later if necessary. */
                object->mc->save_outgoing_msg(object->mc,
                                              object->xid,
                                              MC_RELEASE,
                                              MC_ACK,
                                              lease_id,
                                              lease_id_len,
                                              ack);
            }
        } else {
            ok = FAILURE;
        }
        result = ok;
    } else if((ok == SUCCESS) && (send_ack == FALSE)) {
        /* Send NAK. */
        nak = madcap_msg_nak_new(object->n,
                                 object->ns,
                                 object->xid,
                                 object->data,
                                 object->data_len);
        if(nak != NULL) {
            ok = nak->send(nak);
            if(ok == SUCCESS) {
                /* Save message so it can be resent later if necessary. */
                object->mc->save_outgoing_msg(object->mc,
                                              object->xid,
                                              MC_RELEASE,
                                              MC_NAK,
                                              lease_id,
                                              lease_id_len,
                                              nak);
            }
        } else {
            ok = FAILURE;
        }
        result = ok;
    }

    /* Create error message if something failed. */
    if(ok == FAILURE) {
        if(send_ack == TRUE) {
            mc_error->notify(mc_error, MC_ERR_CREATE_ACK,
                             "madcap_msg_release: Failed to create and send.");
            result = FAILURE;
        } else {
            mc_error->notify(mc_error, MC_ERR_CREATE_NAK,
                             "madcap_msg_release: Failed to create and send.");
            result = FAILURE;
        }
    }

    format->delete(format);

    return result;
}


/*
 * Deletes a RELEASE message.
 * The function deletes a RELEASE message and returnes used resources to the
 * system.
 *
 * @param msg The message that should be deleted.
 * @return 0 on success and -1 on failure.
 */
static int pub_delete(madcap_msg_release *object) {

    free(object->data);
    free(object);

    return SUCCESS;
}


