#ifndef __NET_H__
#define __NET_H__

#include "tim.h"
#include "list.h"
#include "net_socket.h"
#include "madcap_error.h"
#include "debug.h"

/**
 * net.h
 *
 * Implements a network abstraction object. The net object abstracts
 * a packet oriented network, such as a UDP over IP network.
 *
 * Each network endpoint is identified by a net_socket object (see net_socket.h)
 * , which abstracts the socket layer of the underlying OS network
 * implementation.Datagrams are received by registering a callback function
 * with a net_socket, by using the receive method. Datagrams are sent with the
 * send method.
 *
 * Associated with the net object is a timer (see tim.h), which can be used to
 * peridoically (or otherwise timed) call a function. The timer must have been
 * instanciated (with tim_new()) before calling net_new().
 *
 * The net object contains a main method which constitutes the main loop of the
 * program using the net object. This main loop will block, waiting for
 * arriving datagrams or time outs (from the timer). When a datagram in
 * received or the timer expires, the appropriate callback function is called.
 *
 * Example:
 *
 * 1) Set up a net_socket:
 *
 *    net_socket *ns = net_socket_new(listen_ip_addr, listen_port);
 *
 * 2) Register a callback function with receive
 *
 *    int callback(void *obj, char *data, int len) {        
 *       printf("%s: Datagram received with length %d\n", (char *)obj, len);
 *       process_datagram(data, len);
 *    }
 *
 *    ...
 *
 *       net->receive(net, ns, callback, "Callback 1");
 *
 * 3) Initiate the main loop. The callback function "callback" will now be
 *    called with the argument "Callback 1" whenever a datagram is received on
 *    IP address listen_ip_addr, and port listen_port.
 *
 *       net->main(net);
 *
 * 4) To send datagrams, the same net_socket can be used:
 *
 *       ns->set_dest(ns, dest_ipaddr, dest_port);
 *       net->send(net, ns, datagram, datagram_len);
 *
 * Methods:
 *
 * void receive(net *n, net_socket *ns, net_callback nc, void *obj);
 *
 *   Register the callback function nc to be called whenever datagrams
 *   are received on the net_socket ns. obj will be passed as the first
 *   argument to nc when called.
 *
 *   The return value identifies the registration, and can be used to
 *   unregister the callback function (through unreceive).
 *
 * void unreceive(net *n, int id);
 *
 *   Unregisters the callback function identified by id. This identifier
 *   is obtained when invoking the receive method.
 *
 * int send(net *n, net_socket *ns, char *data, int length);
 *
 *   Sends a datagram. The net_socket ns must be set up prior to the call to
 *   send. See the above example for details.
 *
 * int main(net *n);
 *
 *   Constitues the main loop of the program using the net object. A typical
 *   program using a net object will look like:
 *
 *       int main(int argc, char **argv) {
 *          tim *timer;
 *          net *network;
 * 
 *          timer   = tim_new();
 *          network = net_new(tim);
 *
 *          init_timers(timer);
 *          init_network_callbacks(network);
 *          init_everything_else();
 *        
 *          network->main(network);
 *          return 0;
 *       }
 *
 */
 



typedef struct net_peer_ net_peer;

struct net_peer_ {
    unsigned long peer_addr; /* Address of sender. (network byte order) */
    unsigned long dest_addr; /* Destination address used by sender
                                (network byte order). */
    unsigned int  peer_port; /* Port used (host byte order) */
    char ip_addr[17];
};


typedef struct net_ net;
typedef int (*net_callback)(void *obj, char *data, int len, net_peer *np);

struct net_ {
    
    net_peer *sender;  
    int nreg;           
    list *registrations; /* the list of registered callbacks/net_sockets */
    tim *timer;           /* the timer object */
    void (*receive)(net *n, net_socket *s, net_callback nc, void *obj);
    void (*unreceive)(net *n, net_socket *ns);
    int (*send)(net *n, net_socket *s, char *data, int length);
    int (*main)(net *n);
    
    int (* is_self)(net *n, unsigned long addr);
    
    net_socket *(* socket)(net *n);
    
    
};

net *net_new(tim *timer);

#endif /* __NET_H__ */
