#ifndef __NET_SOCKET_H__
#define __NET_SOCKET_H__

#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>

/**
 * net_socket.h
 *
 * Abstracts a network connection. Each network connection must be identified
 * by a net_socket object. The constructor, net_socket_new() takes two arguments:
 * the IP address and the UDP port number on which it will listen for incomming datagrams
 * when registered with a net object (see net.h). Before sending datagrams, the net_socket
 * will have to be set up using the set_dest() method.
 *
 * Methods:
 *
 * int set_dest(net_socket *ns, char *ip_address, int port);
 *
 *   Sets the IP address and UDP port given by the ip_address, port pair to
 *   be the destination for any datagrams sent over this net_socket.
 *
 * int bind(net_socket *ns, char *ip_address, int port);
 *
 *   Binds the net_socket ns to an IP address and UDP port. This means that
 *   datagrams destined for this address/port will be picked up by the
 *   socket. If the IP address is a multicast address (>= 224.0.0.0), the
 *   corresponding multicast group will be joined.
 *
 *   If the net_socket should be bound to any IP address (i.e., any interface),
 *   the value NET_SOCKET_ANY should be used as the ip_address argument.
 *
 * int delete(net_socket *ns);
 *
 *   Deletes the net_socket ns. Closes any sockets associated with the net_socket
 *   and frees any memory previously allocated.
 *
 */


#define NET_SOCKET_ANY NULL

typedef struct net_socket_ net_socket;

struct net_socket_ {

    int socket;
    struct sockaddr_in sa_dest;
    int (*delete)(net_socket *ns);
    int (*set_dest)(net_socket *ns, char *ip_address, int port);
    int (*bind)(net_socket *ns, char *ip_address, int port);
    unsigned long (* name)(net_socket *ns);
    int (*join_multicast_group)(net_socket *ns, unsigned long ip_address,
                                unsigned int ttl);
};

net_socket *net_socket_new();

#endif /* __NET_SOCKET_H__ */
